@@ -9,6 +9,7 @@ import Control.Concurrent (forkFinally, forkIO)
 import qualified Control.Exception as E
 import Control.Monad (forever, void)
 import Data.ByteString (ByteString)
+import qualified Data.List.NonEmpty as NE
 import Network.Socket
 import Network.Socket.ByteString
 
@@ -19,15 +20,15 @@ import Network.Run.Core
 --   server's socket address.
 --   They should be used with 'sendTo'.
 runUDPClient :: HostName -> ServiceName -> (Socket -> SockAddr -> IO a) -> IO a
-runUDPClient host port client = withSocketsDo $ do
-    addr <- resolve Datagram (Just host) port [AI_ADDRCONFIG]
+runUDPClient host port client = do
+    addr <- resolve Datagram (Just host) port [AI_ADDRCONFIG] NE.head
     let sockAddr = addrAddress addr
     E.bracket (openSocket addr) close $ \sock -> client sock sockAddr
 
 -- | Running a UDP server with an open socket in a single Haskell thread.
 runUDPServer :: Maybe HostName -> ServiceName -> (Socket -> IO a) -> IO a
-runUDPServer mhost port server = withSocketsDo $ do
-    addr <- resolve Datagram mhost port [AI_PASSIVE]
+runUDPServer mhost port server = do
+    addr <- resolve Datagram mhost port [AI_PASSIVE] NE.head
     E.bracket (openServerSocket addr) close server
 
 -- | Running a UDP server with a connected socket in each Haskell thread.
@@ -46,19 +47,21 @@ runUDPServerFork (h : hs) port server = do
     mapM_ (forkIO . run) hs
     run h
   where
-    run host = runUDPServer (Just host) port $ \lsock -> forever $ do
-        (bs0, peeraddr) <- recvFrom lsock 2048
-        let family = case peeraddr of
-                SockAddrInet{} -> AF_INET
-                SockAddrInet6{} -> AF_INET6
-                _ -> error "family"
-            hints =
-                defaultHints
-                    { addrSocketType = Datagram
-                    , addrFamily = family
-                    , addrFlags = [AI_PASSIVE]
-                    }
-        addr <- head <$> getAddrInfo (Just hints) Nothing (Just port)
-        s <- openServerSocket addr
-        connect s peeraddr
-        void $ forkFinally (server s bs0) (\_ -> close s)
+    run host = do
+        labelMe $ "UDP server for " ++ h
+        runUDPServer (Just host) port $ \lsock -> forever $ do
+            (bs0, peeraddr) <- recvFrom lsock 2048
+            let family = case peeraddr of
+                    SockAddrInet{} -> AF_INET
+                    SockAddrInet6{} -> AF_INET6
+                    _ -> error "family"
+                hints =
+                    defaultHints
+                        { addrSocketType = Datagram
+                        , addrFamily = family
+                        , addrFlags = [AI_PASSIVE]
+                        }
+            addr <- NE.head <$> getAddrInfo (Just hints) Nothing (Just port)
+            s <- openServerSocket addr
+            connect s peeraddr
+            void $ forkFinally (labelMe "UDP server" >> server s bs0) (\_ -> close s)
